import os
from pathlib import Path

import git

from . import user_repo


def permalink(
    repo: git.Repo | None = None, filepath: str | os.PathLike[str] | None = None
) -> str | None:
    if repo is None:
        repo = git.Repo(search_parent_directories=True)
    user: str | None
    repo_str: str | None
    user, repo_str = user_repo(repo)
    if not (user and repo):
        return None
    sha: str = repo.head.commit.hexsha
    link: str = f"https://github.com/{user}/{repo_str}/tree/{sha}"
    if filepath:
        filepath = Path(filepath).absolute()
        if not filepath.is_relative_to(repo.working_dir):
            return None
        link += f"/{Path(filepath).as_posix()}"



import os
from pathlib import Path
from typing import Any

import neptune
import neptune.common.exceptions
import neptune.utils
import pydantic
import pydantic_settings as ps

from liblaf import cherries


class BackendNeptune(cherries.Backend):
    model_config = ps.SettingsConfigDict(frozen=True, env_prefix="NEPTUNE_")
    monitoring_namespace: str | None = None
    _backend: neptune.Run = pydantic.PrivateAttr()

    @property
    def backend(self) -> str:
        return "neptune"

    @property
    def id(self) -> str:
        return self._backend["sys/id"].fetch()

    @property
    def name(self) -> str:
        return self._backend["sys/name"].fetch()

    @property
    def url(self) -> str:
        return self._backend.get_url()

    def start(self) -> None:
        neptune.common.exceptions.STYLES.update(neptune.common.exceptions.EMPTY_STYLES)
        self._backend = neptune.init_run(monitoring_namespace=self.monitoring_namespace)

    def end(self) -> None:
        self._backend.stop()

    def log_metric(
        self,
        key: str,
        value: float,
        *,
        step: float | None = None,
        timestamp: float | None = None,
        **kwargs,
    ) -> None:
        self._backend[key].append(value, step=step, timestamp=timestamp, **kwargs)

    def log_other(self, key: str, value: Any, **kwargs) -> None:
        value = stringify_unsupported(value)
        self._backend[key].assign(value, **kwargs)

    def upload_file(self, key: str, path: Path, **kwargs) -> None:
        return self._backend[key].upload(str(path), **kwargs)


def stringify_unsupported(value: Any) -> Any:
    if isinstance(value, pydantic.BaseModel):
        return value.model_dump()
    if isinstance(value, os.PathLike):
        return str(value)
    return value



import datetime
import functools
from pathlib import Path
from typing import Any

import pydantic_settings as ps

from liblaf import cherries


class Backend(ps.BaseSettings):
    model_config = ps.SettingsConfigDict(
        frozen=True, env_prefix=cherries.ENV_PREFIX + "DUMMY_"
    )
    enabled: bool = True

    @property
    def backend(self) -> str:
        return "dummy"

    @functools.cached_property
    def entrypoint(self) -> Path:
        return cherries.entrypoint()

    @property
    def id(self) -> str:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError

    @functools.cached_property
    def start_time(self) -> datetime.datetime:
        return datetime.datetime.now().astimezone()

    @property
    def url(self) -> str:
        raise NotImplementedError

    def start(self) -> None: ...
    def end(self) -> None: ...
    def log_metric(
        self,
        key: str,
        value: float,
        *,
        step: float | None = None,
        timestamp: float | None = None,
        **kwargs,
    ) -> None: ...
    def log_other(self, key: str, value: Any, **kwargs) -> None: ...
    def upload_file(self, key: str, path: Path, **kwargs) -> None: ...



import contextlib
import json
import os
import subprocess as sp
from pathlib import Path
from typing import Any

import pydantic
import pydantic_settings as ps
from loguru import logger

from liblaf import cherries


def default_config() -> Path:
    git_root: Path = cherries.git.root()
    for path in [
        git_root / ".config" / "resticprofile.toml",
        git_root / "resticprofile.toml",
    ]:
        if path.exists():
            return path
    return git_root / ".config" / "resticprofile.toml"


class PluginRestic(cherries.Plugin):
    model_config = ps.SettingsConfigDict(env_prefix=cherries.ENV_PREFIX + "RESTIC_")
    config: Path = pydantic.Field(default_factory=default_config)
    name: str | None = None
    dry_run: bool = False

    def _pre_end(self, run: cherries.Experiment) -> None:
        if not self.config.exists():
            logger.warning("configuration file '{}' was not found", self.config)
            return
        args: list[str | os.PathLike[str]] = [
            "resticprofile",
            "--config",
            self.config,
            "backup",
            "--json",
        ]
        if self.name:
            args += ["--name", self.name]
        if self.dry_run:
            args.append("--dry-run")
        args += ["--time", run.start_time.strftime("%Y-%m-%d %H:%M:%S")]
        proc: sp.Popen[str] = sp.Popen(
            args, stdout=sp.PIPE, cwd=cherries.git.root(), text=True
        )
        assert proc.stdout is not None
        for line_ in proc.stdout:
            line: str = line_.rstrip()
            logger.debug("{}", line)
            with contextlib.suppress(json.JSONDecodeError):
                log: dict[str, Any] = json.loads(line)
                if log["message_type"] == "summary":
                    run.log_other("cherries/restic", log)
        returncode: int = proc.wait()
        run.log_other("cherries/restic/returncode", returncode)



from collections.abc import Sequence

from environs import Env

from liblaf import cherries


def start(
    *,
    backend: cherries.Backend | None = None,
    enabled: bool | None = None,
    plugins: Sequence[cherries.Plugin] | None = None,
) -> cherries.Experiment:
    backend = backend or cherries.backend_factory()
    if enabled is None:
        enabled = Env().bool("LIBLAF_CHERRIES_ENABLED", True)
    if plugins is None:
        plugins = cherries.default_plugins()
    exp = cherries.Experiment(backend=backend, enabled=enabled, plugins=plugins)
    exp.start()
    return exp


def end() -> None:
    exp: cherries.Experiment = cherries.current_experiment()
    exp.end()



import subprocess as sp

import git

DEFAULT_COMMIT_MESSAGE: str = """chore(exp): auto commit"""


def commit(message: str = DEFAULT_COMMIT_MESSAGE, *, dry_run: bool = False) -> bool:
    repo = git.Repo(search_parent_directories=True)
    if not repo.is_dirty(untracked_files=True):
        return False
    repo.git.add(all=True, dry_run=dry_run)
    sp.run(["git", "status"], check=False)
    if dry_run:
        return False
    repo.git.commit(message=message)
    return True



import atexit
import dataclasses
import datetime
import os
from collections.abc import Sequence
from pathlib import Path
from typing import Any

from environs import env
from loguru import logger

from liblaf import cherries


@dataclasses.dataclass(kw_only=True)
class Experiment:
    backend: cherries.Backend = dataclasses.field(
        default_factory=cherries.backend_factory
    )
    enabled: bool = dataclasses.field(
        default_factory=lambda: env.bool("LIBLAF_CHERRIES_ENABLED", True)
    )
    plugins: Sequence[cherries.Plugin] = dataclasses.field(
        default_factory=cherries.default_plugins
    )

    @property
    def entrypoint(self) -> Path:
        return self.backend.entrypoint

    @property
    def id(self) -> str:
        return self.backend.id

    @property
    def name(self) -> str:
        return self.backend.name

    @property
    def start_time(self) -> datetime.datetime:
        return self.backend.start_time

    @property
    def url(self) -> str:
        return self.backend.url

    def start(self) -> None:
        if not self.enabled:
            return
        self.plugins = sorted(self.plugins, key=lambda plugin: plugin.priority)
        for plugin in self.plugins:
            plugin.pre_start()
        self.backend.start()
        cherries.set_current_experiment(self)
        for plugin in self.plugins:
            plugin.post_start(self)
        self.log_other("cherries/entrypoint", self.entrypoint)
        self.log_other("cherries/start_time", self.start_time)
        atexit.register(self.end)

    def end(self) -> None:
        if not self.enabled:
            return
        for plugin in reversed(self.plugins):
            plugin.pre_end(self)
        self.backend.end()
        for plugin in reversed(self.plugins):
            plugin.post_end(self)
        self.enabled = False  # prevent `end()` from being called multiple times

    def log_metric(
        self,
        key: str,
        value: float,
        *,
        step: float | None = None,
        timestamp: float | None = None,
        **kwargs,
    ) -> None:
        if not self.enabled:
            return
        logger.opt(depth=1).debug("{}: {}", key, value)
        self.backend.log_metric(key, value, step=step, timestamp=timestamp, **kwargs)

    def log_other(self, key: str, value: Any, **kwargs) -> None:
        if not self.enabled:
            return
        logger.opt(depth=1).info("{}: {}", key, value)
        self.backend.log_other(key, value, **kwargs)

    def upload_file(self, key: str, path: str | os.PathLike[str], **kwargs) -> None:
        if not self.enabled:
            return
        path = Path(path)
        logger.opt(depth=1).info("Uploading file: {}", path)
        self.backend.upload_file(key, path, **kwargs)


_current_experiment: Experiment | None = None


def current_experiment() -> Experiment:
    global _current_experiment  # noqa: PLW0603
    if _current_experiment is None:
        _current_experiment = Experiment()
    return _current_experiment


def set_current_experiment(experiment: Experiment) -> None:
    global _current_experiment  # noqa: PLW0603
    _current_experiment = experiment



import re

import git

GITHUB_URL_PATTERNS: list[str] = [
    r"https://github\.com/(?P<user>[^/]+)/(?P<repo>[^/]+)(?:\.git)?"
]


def user_repo(
    repo: git.Repo | None = None,
) -> tuple[str, str] | tuple[None, None]:
    if repo is None:
        repo = git.Repo(search_parent_directories=True)
    remote: git.Remote = repo.remote()
    for pattern in GITHUB_URL_PATTERNS:
        match: re.Match[str] | None = re.match(pattern, remote.url)
        if not match:
            continue
        user: str = match.group("user")
        repo_name: str = match.group("repo")
        return user, repo_name
    return None, None



import git
import pydantic_settings as ps

from liblaf import cherries


class PluginGit(cherries.Plugin):
    model_config = ps.SettingsConfigDict(env_prefix=cherries.ENV_PREFIX + "GIT_")
    auto_commit: bool = True
    auto_commit_message: str = cherries.git.DEFAULT_COMMIT_MESSAGE

    def _pre_start(self) -> None:
        if self.auto_commit:
            cherries.git.commit(self.auto_commit_message)

    def _post_start(self, run: cherries.Experiment) -> None:
        r = git.Repo(search_parent_directories=True)
        sha: str = r.head.commit.hexsha
        run.log_other("cherries/git/sha", sha)
        if browse := cherries.git.permalink(repo=r):
            run.log_other("cherries/git/browse", browse)
        if entrypoint := cherries.git.permalink(repo=r, filepath=run.entrypoint):
            run.log_other("cherries/git/entrypoint", entrypoint)



from liblaf import cherries


def default_plugins() -> list[cherries.Plugin]:
    return [
        cherries.plugin.PluginLogging(),
        cherries.plugin.PluginGit(),
        cherries.plugin.PluginRestic(),
    ]



import loguru
import pydantic_settings as ps
from loguru import logger

# TODO: fix PLR0402
# make pyright happy
import liblaf.grapes as grapes  # noqa: PLR0402
from liblaf import cherries

DEFAULT_FILTER: "loguru.FilterDict" = {
    "": "INFO",
    "__main__": "TRACE",
    "liblaf": "DEBUG",
}
DEFAULT_FILE_FILTER: "loguru.FilterDict" = {
    **DEFAULT_FILTER,
    "liblaf.cherries": "SUCCESS",
}


class PluginLogging(cherries.Plugin):
    model_config = ps.SettingsConfigDict(env_prefix=cherries.ENV_PREFIX + "LOGGING_")

    def _pre_start(self) -> None:
        grapes.init_logging(
            handlers=[
                grapes.logging.console_handler(),
                grapes.logging.file_handler(),
                grapes.logging.jsonl_handler(),
            ]
        )

    def _pre_end(self, run: cherries.Experiment) -> None:
        logger.complete()
        run.upload_file("cherries/logging/run.log", "run.log")
        run.upload_file("cherries/logging/run.log.jsonl", "run.log.jsonl")



from . import github
from ._commit import DEFAULT_COMMIT_MESSAGE, commit
from ._entrypoint import entrypoint
from ._grapes import GitInfo, info, root, root_safe
from .github import permalink, user_repo

__all__ = [
    "DEFAULT_COMMIT_MESSAGE",
    "GitInfo",
    "commit",
    "entrypoint",
    "github",
    "info",
    "permalink",
    "root",
    "root_safe",
    "user_repo",
]
